import graphene
import pytest
from ...fields import AclField, acl_field_type


class FooEnum(graphene.Enum):
    FOO = "foo"
    BAR = "bar"
    FOOBAR = "foobar"


def classifier(root, info, *args, **kwargs):
    permissions = info.context.jwt.permissions

    if "foo" in permissions:
        return "foo"
    elif "bar" in permissions:
        return "bar"

    return None


def and_or_classifier(root, info, *args, **kwargs):
    permissions = info.context.jwt.permissions

    if "foo" in permissions and "bar" not in permissions:
        return "foo"
    elif "bar" in permissions and "foo" not in permissions:
        return "bar"
    elif "bar" in permissions and "foo" in permissions:
        return ["bar", "foo"]

    return None


def enum_classifier(root, info, foo, *args, **kwargs):
    if foo in [FooEnum.FOO]:
        return foo
    elif foo in [FooEnum.BAR, FooEnum.FOOBAR]:
        return foo


def has_permissions(permissions):
    def validator(root, info, *args, **kwars):
        user_perms = info.context.jwt.permissions
        if any([permission in user_perms for permission in permissions]):
            return True
        raise Exception(f"Not authorized to query field {info.field_name}")

    return validator


def test_AclField_validator_success(mocker):
    field = AclField(
        graphene.String,
        acl_validator=has_permissions(["foo", "bar"]),
        resolver=lambda root, info: "name",
    )

    resolver = field.get_resolver(None)
    mock_info = mocker.MagicMock()
    mock_info.context.jwt.permissions = ["foo"]

    result = resolver(None, mock_info)

    assert result == "name"

    mock_info.context.jwt.permissions = ["bar"]

    result = resolver(None, mock_info)

    assert result == "name"

    mock_info.context.jwt.permissions = []

    try:
        result = resolver(None, mock_info)
        assert False, "Failed to raise expected exception"
    except Exception:
        assert True


def test_AclField_classifier_success(mocker):
    field = AclField(graphene.String, acl_classifier=classifier)

    @field.resolve("foo")
    def foo_resolver(root, info, *args, **kwargs):
        return "foo"

    @field.resolve("bar")
    def bar_resolver(root, info, *args, **kwargs):
        return "bar"

    @field.resolve()
    def default_resolver(root, info, *args, **kwargs):
        return "default"

    resolver = field.get_resolver(None)

    mock_info = mocker.MagicMock()
    mock_info.context.jwt.permissions = ["foo"]

    result = resolver(None, mock_info)

    assert result == "foo"

    mock_info.context.jwt.permissions = ["bar"]

    result = resolver(None, mock_info)

    assert result == "bar"

    mock_info.context.jwt.permissions = []

    result = resolver(None, mock_info)

    assert result == "default"


def test_AclField_multi_route_classifier_success(mocker):
    field = AclField(graphene.String, acl_classifier=and_or_classifier)

    @field.resolve("foo")
    @field.resolve("bar")
    def foo_resolver(root, info, *args, **kwargs):
        return "foo or bar"

    @field.resolve(["foo", "bar"])
    def foo_and_bar_resolver(root, info, *args, **kwargs):
        return "foo and bar"

    @field.resolve()
    def default_resolver(root, info, *args, **kwargs):
        return "default"

    resolver = field.get_resolver(None)

    mock_info = mocker.MagicMock()
    mock_info.context.jwt.permissions = ["foo"]

    result = resolver(None, mock_info)

    assert (
        result == "foo or bar"
    ), "Failed to resolve stacked resolver decorators"

    mock_info.context.jwt.permissions = ["bar"]

    result = resolver(None, mock_info)

    assert (
        result == "foo or bar"
    ), "Failed to resolve stacked resolver decorators"

    mock_info.context.jwt.permissions = ["bar", "foo"]

    result = resolver(None, mock_info)

    assert result == (
        "foo and bar"
    ), "Failed to resolve ANDed route keys for 1:1 permissions"

    mock_info.context.jwt.permissions = ["bar", "foo", "foobar"]

    result = resolver(None, mock_info)

    assert result == (
        "foo and bar"
    ), "Failed to resolve ANDed route keys for extra permissions"

    mock_info.context.jwt.permissions = []

    result = resolver(None, mock_info)

    assert result == "default"


def test_AclField_enum_classifier_success(mocker):
    field = AclField(
        graphene.String,
        foo=graphene.Argument(FooEnum),
        acl_classifier=enum_classifier,
    )

    @field.resolve(FooEnum.FOO)
    def foo_resolver(root, info, *args, **kwargs):
        return "foo"

    @field.resolve(FooEnum.BAR)
    def bar_resolver(root, info, *args, **kwargs):
        return "bar"

    @field.resolve(FooEnum.FOOBAR.value)
    def default_resolver(root, info, *args, **kwargs):
        return "foobar"

    resolver = field.get_resolver(None)

    mock_info = mocker.MagicMock()

    result = resolver(None, mock_info, "foo")

    assert result == "foo"

    result = resolver(None, mock_info, FooEnum.FOO)

    assert result == "foo"

    result = resolver(None, mock_info, FooEnum.FOO.value)

    assert result == "foo"

    result = resolver(None, mock_info, "bar")

    assert result == "bar"

    result = resolver(None, mock_info, FooEnum.BAR)

    assert result == "bar"

    result = resolver(None, mock_info, "foobar")

    assert result == "foobar"

    result = resolver(None, mock_info, FooEnum.FOOBAR.value)

    assert result == "foobar"

    result = resolver(None, mock_info, "RABOOF")

    assert result is None


class CustomField(graphene.Field):
    def _resolver(self, parent_resolver, root, info, *args, **kwargs):
        response = parent_resolver(root, info, *args, **kwargs)
        if not response:
            return response
        return response + "bar"

    def get_resolver(self, parent_resolver):
        from functools import partial

        resolver = super(CustomField, self).get_resolver(parent_resolver)
        return partial(self._resolver, resolver)


def test_acl_field_type(mocker):
    AclCustomField = acl_field_type("AclCustomField", CustomField)

    field = AclCustomField(
        graphene.String,
        acl_validator=has_permissions(["foo"]),
        resolver=lambda root, info: "foo",
    )

    resolver = field.get_resolver(None)
    mock_info = mocker.MagicMock()
    mock_info.context.jwt.permissions = ["foo"]

    result = resolver(None, mock_info)

    assert result == "foobar"

    mock_info.context.jwt.permissions = ["bar"]

    with pytest.raises(Exception):
        resolver(None, mock_info)



import graphene
import pytest
import asyncio
from ...fields import AclField


async def async_classifier(root, info, *args, **kwargs):
    permissions = info.context.jwt.permissions
    await asyncio.sleep(0.1)

    if "foo" in permissions:
        return "foo"
    elif "bar" in permissions:
        return "bar"

    return None


def classifier(root, info, *args, **kwargs):
    permissions = info.context.jwt.permissions

    if "foo" in permissions:
        return "foo"
    elif "bar" in permissions:
        return "bar"

    return None


async_classifier_field = AclField(
    graphene.String, acl_classifier=async_classifier
)
sync_classifier_field = AclField(graphene.String, acl_classifier=classifier)


@async_classifier_field.resolve("foo")
@sync_classifier_field.resolve("foo")
async def foo_resolver(root, info, *args, **kwargs):
    await asyncio.sleep(0.1)
    return "foo"


@async_classifier_field.resolve("bar")
@sync_classifier_field.resolve("bar")
def bar_resolver(root, info, *args, **kwargs):
    return "bar"


@async_classifier_field.resolve()
@sync_classifier_field.resolve()
def default_resolver(root, info, *args, **kwargs):
    return "default"


@pytest.mark.asyncio
async def test_AclField_async_classifier_success(mocker):
    info = mocker.MagicMock()

    resolver = async_classifier_field.get_resolver(None)

    info.context.jwt.permissions = ["foo"]

    result = await resolver(None, info)

    assert result == "foo"

    info.context.jwt.permissions = ["bar"]

    result = await resolver(None, info)

    assert result == "bar"

    info.context.jwt.permissions = []

    result = await resolver(None, info)

    assert result == "default"


@pytest.mark.asyncio
async def test_AclField_sync_classifier_success(mocker):
    info = mocker.MagicMock()
    resolver = sync_classifier_field.get_resolver(None)

    info.context.jwt.permissions = ["foo"]

    result = await resolver(None, info)

    assert result == "foo"

    info.context.jwt.permissions = ["bar"]

    result = resolver(None, info)

    assert result == "bar"

    info.context.jwt.permissions = []

    result = resolver(None, info)

    assert result == "default"



from logging import getLogger
from promise import Promise
from enum import Enum

logger = getLogger(__name__)


def _hash_keys(keys=None):
    if keys:
        if type(keys) in [list, tuple]:
            sorted_keys = sorted(keys)
            return tuple(sorted_keys)
        return str(keys)
    return keys


def _default_handler(root, info, *args, **kwargs):
    return None


class ResolverRouter(object):
    def __init__(self, classifier=None):
        self._handlers = {}
        self._classifier = classifier
        self._default_handler = None

    def register(self, handler, key=None):
        hashable_key = _hash_keys(key)
        if hashable_key:
            if isinstance(key, Enum):
                self._handlers[key.value] = handler
                self._handlers[key.name] = handler
                self._handlers[str(key)] = handler
            else:
                self._handlers[hashable_key] = handler
        else:
            self._default_handler = handler

    def _get_handler(self, root, info, *args, routing_key=None, **kwargs):
        handler = None

        if routing_key:
            routing_key_hashable = _hash_keys(routing_key)
            handler = self._handlers.get(routing_key_hashable, None)
            if not handler:
                logger.error(
                    "Missing resolver for {}.{} classifier {}".format(
                        self.__class__, info.field_name, routing_key_hashable
                    )
                )
        return handler or self._default_handler

    def _resolver(
        self, routing_key, parent_resolver, root, info, *args, **kwargs
    ):
        handler = None

        if routing_key:
            handler = self._get_handler(
                root, info, *args, routing_key=routing_key, **kwargs
            )

        if not handler:
            handler = self._default_handler or parent_resolver

        if handler:
            if Promise.is_thenable(handler):
                return Promise.resolve(handler(root, info, *args, **kwargs))

            return handler(root, info, *args, **kwargs)
        return None

    def resolver(self, parent_resolver, root, info, *args, **kwargs):
        routing_key = None
        if self._classifier:
            routing_key = self._classifier(root, info, *args, **kwargs)
            if Promise.is_thenable(routing_key):
                return Promise.resolve(routing_key).then(
                    lambda key: self._resolver(
                        key, parent_resolver, root, info, *args, **kwargs
                    )
                )
        return self._resolver(
            routing_key, parent_resolver, root, info, *args, **kwargs
        )



from logging import getLogger
from graphene import Field
from functools import partial, wraps
from promise import Promise
from .routing import ResolverRouter

logger = getLogger(__name__)


class AclField(Field):
    def __init__(
        self, *args, acl_validator=None, acl_classifier=None, **kwargs
    ):
        self._acl_validator = acl_validator
        self._acl_router = ResolverRouter(acl_classifier)
        super(AclField, self).__init__(*args, **kwargs)

    def get_resolver(self, parent_resolver):
        super_resolver = super(AclField, self).get_resolver(parent_resolver)
        resolver = partial(
            self._acl_router.resolver, super_resolver or self.resolver
        )

        @wraps(resolver)
        def _resolver(root, info, *args, **kwargs):
            if self._acl_validator(root, info, *args, **kwargs):
                result = resolver(root, info, *args, **kwargs)
                if Promise.is_thenable(result):
                    return Promise.resolve(result)
                return result

        if self._acl_validator:
            return _resolver
        else:
            return resolver

    def resolve(self, route_key=None):
        def decorator(func):
            self._acl_router.register(func, route_key)
            return func

        return decorator


def acl_field_type(name, WrappedField):
    def get_resolver(self, parent_resolver):
        acl_resolver = AclField.get_resolver(self, parent_resolver)
        return WrappedField.get_resolver(self, acl_resolver)

    return type(name, (WrappedField, AclField), {"get_resolver": get_resolver})



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graphene_acl', 'graphene_acl.tests', 'graphene_acl.tests.unit']

package_data = \
{'': ['*']}

install_requires = \
['graphene>=2.1,<3.0']

setup_kwargs = {
    'name': 'graphene-acl',
    'version': '1.1.1',
    'description': 'Graphene Field ACL',
    'long_description': "# graphene-acl\n\nThe motivation for this library is to simplify access control protection for Graphene Fields. A common approach to ACL protection is through the use of a reusable permissions validation decorator. The problem is this is cumbersome for Graphene Fields that use the standard resolver. You are forced to write an unnecessary resolver function just to annotate it with your permissions validator. The second cumbersome problem this library addresses is ACL role based resolvers. Depending on the users role you might want to perform different business logic in order to retrieve the data they requested for a Graphene Field.\n\n## Installation\n\n```bash\n$ pip install graphene-acl\n```\n\n## Usage\n\n### acl_classifier\n\nThe purpose of the classifier is to return a route key that will be used to determine which resolver function is used for resolving the field. The classifier function has access to all the arguments from the field resolver and can be excuted synchonously or asynchronously.\n\n### acl_validator\n\nThe purpose of the validator is to authorize access to the field. This validation will occurr before classification routing happens. If authorization validation is different per classification route then you should not use this validator to enforce authorization access. Instead you should authorize at the specific classifier resolver definition.\n\n### Example\n\n```python\nfrom graphene_acl import AclField\nimport graphene\n\ndef classifier(root, info, *args, **kwargs):\n    permissions = info.context.jwt.permissions\n\n    if 'admin' in permissions:\n        return 'admin'\n    if 'perm1' in permissions and not 'perm2' in permissions:\n        return 'perm1'\n    if 'perm2' in permissions and not 'perm1' in permissions:\n        return 'perm2'\n    if 'perm2' in permissions and 'perm1' in permissions:\n        return ['perm1', 'perm2']\n    if 'perm3' in permissions:\n        return 'perm3'\n    if 'perm4' in permissions:\n        return 'perm4'\n    return None\n\ndef has_permissions(permissions):\n    def validator(root, info, *args, **kwars):\n        if (any([permission in info.context.jwt.permissions for permission in permissions])):\n            return True\n        raise AuthorizationError(f'Not authorized to query field {info.field_name}')\n\n    return validator\n\nclass Foo(graphene.ObjectType):\n    # Demonstrates simple routing based on an Admin classifier route key\n    admin_field = AclField(graphene.String, acl_classifier=classifier)\n    restricted_name = AclField(graphene.String, acl_validator=has_permissions(['foo:name:read', 'admin']))\n\n    tenant_field = AclField(graphene.String, acl_classifier=classifier)\n\n@Foo.admin_field.resolve('admin')\ndef resolve_admin_field(root, info, *args, **kwargs):\n    pass\n\n@Foo.admin_field.resolve()\ndef resolve_default_admin_field(root, info, *args, **kwargs):\n    raise Error('Not Authorized')\n\n@Foo.tenant_field.resolve('perm1')\ndef resolve_perm1_field(root, info, *args, **kwargs):\n    pass\n\n@Foo.tenant_field.resolve('perm2')\ndef resolve_perm2_field(root, info, *args, **kwargs):\n    pass\n\n@Foo.tenant_field.resolve(['perm1', 'perm2'])\ndef resolve_perm1_and_perm2_field(root, info, *args, **kwargs):\n    # Uses sorted() + tuple hashing to register function\n    pass\n\n@Foo.tenant_field.resolve('perm3')\n@Foo.tenant_field.resolve('perm4')\ndef resolve_perm3_or_perm4_field(root, info, *args, **kwargs):\n    # Registers function for both 'perm3' and 'perm4' route keys\n    pass\n```\n\n**ACL Connection Fields**\n\n```python\nfrom graphene_django.filter import DjangoFilterConnectionField\nfrom graphene_acl import acl_field_type\n\nBarConnectionField = acl_field_type('BarConnectionField', DjangoFilterConnectionField)\n\nclass Foo(graphene.ObjectType):\n    bar = BarConnectionField(MyNode, acl_permissions=has_permission('FOO'))\n\n```\n\n## Development\n\n### First time setup\n\n-   Install Precommit hooks\n    -   `brew install pre-commit && pre-commit install && pre-commit install --install-hooks`\n-   Install poetry: https://github.com/sdispater/poetry#installation\n    -   `curl -sSL https://raw.githubusercontent.com/sdispater/poetry/master/get-poetry.py | python`\n-   Install dependencies\n    -   `poetry install`\n",
    'author': 'Nick Harris',
    'author_email': 'nick.harris@cybergrx.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CyberGRX/graphene-acl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
